@@ -1,5 +1,5 @@
-#include <license.hunspell>
-#include <license.myspell>
+#include "license.hunspell"
+#include "license.myspell"
 
 #include <cctype>
 #include <cstring>
@@ -50,7 +50,7 @@ PfxEntry::~PfxEntry()
     strip = NULL;
     if (utf8) {
         for (int i = 0; i < 8; i++) {
-            if (conds.utf8.wchars[i]) free(conds.utf8.wchars[i]);  
+            if (conds.utf8.wchars[i]) free(conds.utf8.wchars[i]);
         }
     }
     if (morphcode) free(morphcode);
@@ -61,32 +61,25 @@ PfxEntry::~PfxEntry()
 char * PfxEntry::add(const char * word, int len)
 {
     int			cond;
-    char	        tword[MAXWORDLEN+1];
+    char	        tword[MAXWORDUTF8LEN + 4];
 
-     /* make sure all conditions match */
-     if ((len > stripl) && (len >= numconds)) {
-            unsigned char * cp = (unsigned char *) word;
-            for (cond = 0;  cond < numconds;  cond++) {
-	       if ((conds.base[*cp++] & (1 << cond)) == 0) // XXX UTF-8???
-	          break;
-            }
-            if (cond >= numconds) {
-	      /* we have a match so add prefix */
-              int tlen = 0;
+    if ((len > stripl) && (len >= numconds) && test_condition(word) &&
+       (!stripl || (strncmp(word, strip, stripl) == 0)) && 
+       ((MAXWORDUTF8LEN + 4) > (len + appndl - stripl))) {
+    /* we have a match so add prefix */
+              char * pp = tword;
               if (appndl) {
-	          strcpy(tword,appnd);
-                  tlen += appndl;
-               } 
-               char * pp = tword + tlen;
+                  strcpy(tword,appnd);
+                  pp += appndl;
+               }
                strcpy(pp, (word + stripl));
                return mystrdup(tword);
-	    }
      }
      return NULL;    
 }
 
 
-inline int PfxEntry::test_condition(char * st)
+inline int PfxEntry::test_condition(const char * st)
 {
     int cond;
     unsigned char * cp = (unsigned char *)st;
@@ -133,7 +126,7 @@ struct hentry * PfxEntry::check(const char * word, int len, char in_compound, co
     int	                tmpl;   // length of tmpword
     struct hentry *     he;     // hash entry of root word or NULL
     unsigned char *	cp;		
-    char	        tmpword[MAXWORDLEN+1];
+    char	        tmpword[MAXWORDUTF8LEN + 4];
 
     // on entry prefix is 0 length or already matches the beginning of the word.
     // So if the remaining root word has positive length
@@ -163,10 +156,13 @@ struct hentry * PfxEntry::check(const char * word, int len, char in_compound, co
 		if ((he = pmyMgr->lookup(tmpword)) != NULL) {
                    do {
 		      if (TESTAFF(he->astr, aflag, he->alen) &&
+                        // forbid single prefixes with pseudoroot flag
+                        ! TESTAFF(contclass, pmyMgr->get_pseudoroot(), contclasslen) &&
+                        // needflag
                         ((!needflag) || TESTAFF(he->astr, needflag, he->alen) ||
                          (contclass && TESTAFF(contclass, needflag, contclasslen))))
                             return he;
-                   } while ((he = he->next_homonym)); // check homonyms (HU)
+                   } while ((he = he->next_homonym)); // check homonyms
 		}
                 
 		// prefix matched but no root word was found 
@@ -192,7 +188,7 @@ struct hentry * PfxEntry::check_twosfx(const char * word, int len,
     int	                tmpl;   // length of tmpword
     struct hentry *     he;     // hash entry of root word or NULL
     unsigned char *	cp;		
-    char	        tmpword[MAXWORDLEN+1];
+    char	        tmpword[MAXWORDUTF8LEN + 4];
 
     // on entry prefix is 0 length or already matches the beginning of the word.
     // So if the remaining root word has positive length
@@ -241,7 +237,7 @@ char * PfxEntry::check_twosfx_morph(const char * word, int len,
     int			cond;	// condition number being examined
     int	                tmpl;   // length of tmpword
     unsigned char *	cp;		
-    char	        tmpword[MAXWORDLEN+1];
+    char	        tmpword[MAXWORDUTF8LEN + 4];
 
     // on entry prefix is 0 length or already matches the beginning of the word.
     // So if the remaining root word has positive length
@@ -289,7 +285,7 @@ char * PfxEntry::check_morph(const char * word, int len, char in_compound, const
     int	                tmpl;   // length of tmpword
     struct hentry *     he;     // hash entry of root word or NULL
     unsigned char *	cp;		
-    char	        tmpword[MAXWORDLEN+1];
+    char	        tmpword[MAXWORDUTF8LEN + 4];
     char                result[MAXLNLEN];
     char * st;
     
@@ -323,6 +319,9 @@ char * PfxEntry::check_morph(const char * word, int len, char in_compound, const
 		if ((he = pmyMgr->lookup(tmpword)) != NULL) {
                     do {
 		      if (TESTAFF(he->astr, aflag, he->alen) &&
+                        // forbid single prefixes with pseudoroot flag
+                        ! TESTAFF(contclass, pmyMgr->get_pseudoroot(), contclasslen) &&
+                        // needflag
                         ((!needflag) || TESTAFF(he->astr, needflag, he->alen) ||
                          (contclass && TESTAFF(contclass, needflag, contclasslen)))) {
                             if (morphcode) strcat(result, morphcode); else strcat(result,getKey());
@@ -403,35 +402,26 @@ SfxEntry::~SfxEntry()
 char * SfxEntry::add(const char * word, int len)
 {
     int			cond;
-    char	        tword[MAXWORDLEN+1];
+    char	        tword[MAXWORDUTF8LEN + 4];
 
      /* make sure all conditions match */
-     if ((len > stripl) && (len >= numconds)) {
-            unsigned char * cp = (unsigned char *) (word + len);
-            for (cond = numconds; --cond >=0; ) {
-	       if ((conds.base[*--cp] & (1 << cond)) == 0)
-	          break;
-            }
-            if (cond < 0) {
+     if ((len > stripl) && (len >= numconds) && test_condition(word + len, word) &&
+        (!stripl || (strcmp(word + len - stripl, strip) == 0)) &&
+        ((MAXWORDUTF8LEN + 4) > (len + appndl - stripl))) {
 	      /* we have a match so add suffix */
               strcpy(tword,word);
-              int tlen = len;
-              if (stripl) {
-		 tlen -= stripl;
-              }
-              char * pp = (tword + tlen);
               if (appndl) {
-	          strcpy(pp,appnd);
-                  tlen += appndl;
-	      } else *pp = '\0';
-               return mystrdup(tword);
-	    }
+                  strcpy(tword + len - stripl, appnd);
+              } else {
+                  *(tword + len - stripl) = '\0';
+              }
+              return mystrdup(tword);
      }
      return NULL;
 }
 
 
-inline int SfxEntry::test_condition(char * st, char * beg)
+inline int SfxEntry::test_condition(const char * st, const char * beg)
 {
     int cond;
     unsigned char * cp = (unsigned char *) st;
@@ -482,14 +472,13 @@ struct hentry * SfxEntry::check(const char * word, int len, int optflags,
     int			cond;		 // condition beng examined
     struct hentry *     he;              // hash entry pointer
     unsigned char *	cp;
-    char	        tmpword[MAXWORDLEN+1];
+    char	        tmpword[MAXWORDUTF8LEN + 4];
     PfxEntry* ep = (PfxEntry *) ppfx;
 
-
     // if this suffix is being cross checked with a prefix
     // but it does not support cross products skip it
 
-    if ((optflags & XPRODUCT) != 0 &&  (xpflg & XPRODUCT) == 0)
+    if (((optflags & XPRODUCT) != 0) && ((xpflg & XPRODUCT) == 0))
         return NULL;
 
     // upon entry suffix is 0 length or already matches the end of the word.
@@ -523,13 +512,16 @@ struct hentry * SfxEntry::check(const char * word, int len, int optflags,
             // root word in the dictionary
 
 	    if (test_condition((char *) cp, (char *) tmpword)) {
+
 #ifdef SZOSZABLYA_POSSIBLE_ROOTS
 		fprintf(stdout,"%s %s %c\n", word, tmpword, aflag);
 #endif
 	        if ((he = pmyMgr->lookup(tmpword)) != NULL) {
                     do {
-                        if ((TESTAFF(he->astr, aflag, he->alen) || (ep && ep->getCont() && TESTAFF(ep->getCont(), aflag, ep->getContLen()))) && 
-                            ((optflags & XPRODUCT) == 0 || 
+                        // check conditional suffix (enabled by prefix)
+                        if ((TESTAFF(he->astr, aflag, he->alen) || (ep && ep->getCont() &&
+                                    TESTAFF(ep->getCont(), aflag, ep->getContLen()))) && 
+                            (((optflags & XPRODUCT) == 0) || 
                             TESTAFF(he->astr, ep->getFlag(), he->alen) ||
                              // enabled by prefix
                             ((contclass) && TESTAFF(contclass, ep->getFlag(), contclasslen))
@@ -544,10 +536,11 @@ struct hentry * SfxEntry::check(const char * word, int len, int optflags,
                               ((contclass) && TESTAFF(contclass, needflag, contclasslen)))
                             )
                         ) return he;
-                    } while ((he = he->next_homonym)); // check homonyms (HU)
+                    } while ((he = he->next_homonym)); // check homonyms
 
-	    // store resulting root in wlst (see SuffixMgr:suggest_pos_stems)
-	
+                // obsolote stemming code (used only by the 
+                // experimental SuffixMgr:suggest_pos_stems)
+	        // store resulting root in wlst
 		} else if ((wlst) && (*ns < maxSug) && 
 		    TESTAFF("ABCUVWKLMQRSTqrst", aflag , 17)) { // XXX (HU) productive suffixes
 		    int cwrd = 1;
@@ -576,7 +569,7 @@ struct hentry * SfxEntry::check_twosfx(const char * word, int len, int optflags,
     int			cond;		 // condition beng examined
     struct hentry *     he;              // hash entry pointer
     unsigned char *	cp;
-    char	        tmpword[MAXWORDLEN+1];
+    char	        tmpword[MAXWORDUTF8LEN + 4];
     PfxEntry* ep = (PfxEntry *) ppfx;
 
 
@@ -616,7 +609,7 @@ struct hentry * SfxEntry::check_twosfx(const char * word, int len, int optflags,
 
 	    if (test_condition((char *) cp, (char *) tmpword)) {
                 if (ppfx) {
-                    // handle confixum (HU)
+                    // handle conditional suffix
                     if ((contclass) && TESTAFF(contclass, ep->getFlag(), contclasslen)) 
                         he = pmyMgr->suffix_check(tmpword, tmpl, 0, NULL, NULL, 0, NULL, (FLAG) aflag, needflag);
                     else
@@ -638,7 +631,7 @@ char * SfxEntry::check_twosfx_morph(const char * word, int len, int optflags,
     int	                tmpl;		 // length of tmpword 
     int			cond;		 // condition beng examined
     unsigned char *	cp;
-    char	        tmpword[MAXWORDLEN+1];
+    char	        tmpword[MAXWORDUTF8LEN + 4];
     PfxEntry* ep = (PfxEntry *) ppfx;
     char * st;
 
@@ -682,15 +675,12 @@ char * SfxEntry::check_twosfx_morph(const char * word, int len, int optflags,
 
 	    if (test_condition((char *) cp, (char *) tmpword)) {
                 if (ppfx) {
-                    // handle confixum (HU)
+                    // handle conditional suffix
                     if ((contclass) && TESTAFF(contclass, ep->getFlag(), contclasslen)) {
                         st = pmyMgr->suffix_check_morph(tmpword, tmpl, 0, NULL, aflag, needflag);
                         if (st) {
                             if (((PfxEntry *) ppfx)->getMorph()) {
                                 strcat(result, ((PfxEntry *) ppfx)->getMorph());
-                            } else {
-                                strcat(result,((PfxEntry *) ppfx)->getKey());
-                                strcat(result,"[PREFIX]");
                             }
                             strcat(result,st);
                             free(st);
@@ -729,7 +719,7 @@ struct hentry * SfxEntry::get_next_homonym(struct hentry * he, int optflags, Aff
         if ((TESTAFF(he->astr, aflag, he->alen) || (ep && ep->getCont() && TESTAFF(ep->getCont(), aflag, ep->getContLen()))) && 
                             ((optflags & XPRODUCT) == 0 || 
                             TESTAFF(he->astr, ep->getFlag(), he->alen) ||
-                             // handle confixum (HU)
+                             // handle conditional suffix
                             ((contclass) && TESTAFF(contclass, ep->getFlag(), contclasslen))
                             ) &&
                             // handle cont. class
