@@ -1,9 +1,5 @@
-#include <license.hunspell>
-#include <license.myspell>
-
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include "license.hunspell"
+#include "license.myspell"
 
 #include <cstring>
 #include <cstdlib>
@@ -21,6 +17,7 @@ Hunspell::Hunspell(const char * affpath, const char * dpath)
     csconv = NULL;
     utfconv = NULL;
     utf8 = 0;
+    complexprefixes = 0;
 
     /* first set up the hash manager */
     pHMgr = new HashMgr(dpath, affpath);
@@ -35,11 +32,10 @@ Hunspell::Hunspell(const char * affpath, const char * dpath)
     encoding = pAMgr->get_encoding();
     csconv = get_current_cs(encoding);
     langnum = pAMgr->get_langnum();
-    if (pAMgr->is_utf8()) {
-        utfconv = pAMgr->get_utf_conv();
-        utf8 = 1;
-    }
-    
+    utf8 = pAMgr->get_utf8();
+    utfconv = pAMgr->get_utf_conv();
+    complexprefixes = pAMgr->get_complexprefixes();
+
     /* and finally set up the suggestion manager */
     pSMgr = new SuggestMgr(try_string, MAXSUGGESTION, pAMgr);
     if (try_string) free(try_string);
@@ -88,10 +84,10 @@ int Hunspell::cleanword2(char * dest, const char * src,
        (*pabbrev)++;
    }
    
-   // if no characters are left it can't be an abbreviation and can't be capitalized
+   // if no characters are left it can't be capitalized
    if (nl <= 0) { 
        *pcaptype = NOCAP;
-       *pabbrev = 0;
+//       *pabbrev = 0;
        *p = '\0';
        return 0;
    }
@@ -115,13 +111,15 @@ int Hunspell::cleanword2(char * dest, const char * src,
    } else {
       unsigned short idx;
       *nc = u8_u16(dest_utf, MAXWORDLEN, (const char *) q);
+      // don't check too long words
+      if (*nc >= MAXWORDLEN) return 0;
       *nc -= *pabbrev;
       for (int i = 0; i < *nc; i++) {
          idx = (dest_utf[i].h << 8) + dest_utf[i].l;
          if (idx != utfconv[idx].clower) ncap++;
          if (utfconv[idx].cupper == utfconv[idx].clower) nneutral++;
       }
-      u16_u8(dest, MAXWORDLEN, dest_utf, *nc);
+      u16_u8(dest, MAXWORDUTF8LEN, dest_utf, *nc);
       if (ncap) {
          idx = (dest_utf[0].h << 8) + dest_utf[0].l;
          firstcap = (idx != utfconv[idx].clower);
@@ -137,8 +135,7 @@ int Hunspell::cleanword2(char * dest, const char * src,
         *pcaptype = ALLCAP;
   } else {
         *pcaptype = HUHCAP;
-  }
-  
+  }  
   return strlen(dest);
 } 
 
@@ -160,10 +157,10 @@ int Hunspell::cleanword(char * dest, const char * src,
        (*pabbrev)++;
    }
    
-   // if no characters are left it can't be an abbreviation and can't be capitalized
+   // if no characters are left it can't be capitalized
    if (nl <= 0) { 
        *pcaptype = NOCAP;
-       *pabbrev = 0;
+//       *pabbrev = 0;
        *p = '\0';
        return 0;
    }
@@ -193,7 +190,7 @@ int Hunspell::cleanword(char * dest, const char * src,
          if (idx != utfconv[idx].clower) ncap++;
          if (utfconv[idx].cupper == utfconv[idx].clower) nneutral++;
       }
-      u16_u8(dest, MAXWORDLEN, t, nc);
+      u16_u8(dest, MAXWORDUTF8LEN, t, nc);
       if (ncap) {
          idx = (t[0].h << 8) + t[0].l;
          firstcap = (idx != utfconv[idx].clower);
@@ -228,7 +225,7 @@ void Hunspell::mkallcap(char * p)
             u[i].l = (unsigned char) (utfconv[idx].cupper & 0x00FF);
          }
       }
-      u16_u8(p, MAXWORDLEN, u, nc);
+      u16_u8(p, MAXWORDUTF8LEN, u, nc);
   } else {
     while (*p != '\0') { 
         *p = csconv[((unsigned char) *p)].cupper;
@@ -248,7 +245,7 @@ int Hunspell::mkallcap2(char * p, w_char * u, int nc)
             u[i].l = (unsigned char) (utfconv[idx].cupper & 0x00FF);
          }
       }
-      u16_u8(p, MAXWORDLEN, u, nc);
+      u16_u8(p, MAXWORDUTF8LEN, u, nc);
       return strlen(p);  
   } else {
     while (*p != '\0') { 
@@ -279,7 +276,7 @@ int Hunspell::mkallsmall2(char * p, w_char * u, int nc)
             u[i].l = (unsigned char) (utfconv[idx].clower & 0x00FF);
          }
       }
-      u16_u8(p, MAXWORDLEN, u, nc);
+      u16_u8(p, MAXWORDUTF8LEN, u, nc);
       return strlen(p);
   } else {
     while (*p != '\0') { 
@@ -318,49 +315,49 @@ hentry * Hunspell::spellsharps(char * base, char * pos, int n, int repnum, char
     return NULL;
 }
 
+int Hunspell::is_keepcase(const hentry * rv) {
+    return pAMgr && rv->astr && pAMgr->get_keepcase() &&
+        TESTAFF(rv->astr, pAMgr->get_keepcase(), rv->alen);
+}
+
 int Hunspell::spell(const char * word)
 {
   struct hentry * rv=NULL;
-  char cw[MAXWORDLEN+1];
-  char wspace[MAXWORDLEN+1];
-  w_char unicw[MAXWORDLEN];
-  int nc = strlen(word);
-  int wl2;
-  
   // need larger vector. For example, Turkish capital letter I converted a
   // 2-byte UTF-8 character (dotless i) by mkallsmall.
-  if (nc > ((MAXWORDLEN - 1) / 3)) return 0;
-  
+  char cw[MAXWORDUTF8LEN + 4];
+  char wspace[MAXWORDUTF8LEN + 4];
+  w_char unicw[MAXWORDLEN + 1];
+  int nc = strlen(word);
+  int wl2;
+  if (utf8) {
+    if (nc >= MAXWORDUTF8LEN) return 0;
+  } else {
+    if (nc >= MAXWORDLEN) return 0;
+  }
   int captype = 0;
   int abbv = 0;
   int wl = cleanword2(cw, word, unicw, &nc, &captype, &abbv);
+
   if (wl == 0) return 1;
 
-  // LANG_hu section: numbers with suffixes
-  if (langnum == LANG_hu) {
-    int n = 0;
-    int n2 = 0;
-    int n3 = 0;
-    // test numbers and 56%-kal, 45%-kal 50-50%, 11.20-kor, but
-    // numbers with two, or more dots are bad (for example 2003.07.22.)
-    while ((n < wl) && 
-      (((cw[n] <= '9') && (cw[n] >= '0')) || (((cw[n] == '.') || (cw[n] == ',')) && (n > 0)))) {
-        n++;
-        if ((cw[n] == '.') || (cw[n] == ',')) {
-            if (((n2 == 0) && (n > 3)) || 
-                ((n2 > 0) && ((cw[n-1] == '.') || (cw[n-1] == ',')))) break;
-            n2++;
-            n3 = n;
-        }
-    }
+  // allow numbers with dots and commas (but forbid double separators: "..", ",," etc.)
+  enum { NBEGIN, NNUM, NSEP };
+  int nstate = NBEGIN;
+  int i;
+  
+  for (i = 0; (i < wl) && 
+      (((cw[i] <= '9') && (cw[i] >= '0') && (nstate = NNUM)) ||
+          ((nstate == NNUM) && ((cw[i] == ',') ||
+              (cw[i] == '.') || (cw[i] == '-')) && (nstate = NSEP))); i++);     
+  if ((i == wl) && (nstate == NNUM)) return 1;
 
-    // numbers with 3, or more digits (HU)
-    if ((n == wl) && (n3 > 0) && (n - n3 > 3)) return 0;
-    if ((n == wl) || (((cw[n]=='%') || (cw[n]=='')) && check(cw+n))) return 1;
-    //  if ((cw[n]=='.') && check(cw + n + 1)) return 1;
+  // LANG_hu section: number(s) + (percent or degree) with suffixes
+  if (langnum == LANG_hu) {
+    if ((nstate == NNUM) && ((cw[i] == '%') || (cw[i] == '')) && check(cw + i)) return 1;
   }
   // END of LANG_hu section
-  
+
   switch(captype) {
      case HUHCAP: 
      case NOCAP: { 
@@ -388,7 +385,7 @@ int Hunspell::spell(const char * word)
                         (!utf8 && strchr(cw, '')))) break;
                     // LANG_de section: German Sharp S ()
                     if ((langnum == LANG_de) && strstr(cw, "SS")) {
-                        char tmpword[MAXWORDLEN];
+                        char tmpword[MAXWORDUTF8LEN];
                         wl = mkallsmall2(cw, unicw, nc);
                         memcpy(wspace,cw,(wl+1));
                         rv = spellsharps(wspace, wspace, 0, 0, tmpword);
@@ -415,19 +412,21 @@ int Hunspell::spell(const char * word)
                      wl = mkallsmall2(cw, unicw, nc);
                      memcpy(wspace,cw,(wl+1));                     
                      rv = check(wspace);
-                     if (!rv) {
+                     if (!rv || is_keepcase(rv)) {
                         wl2 = mkinitcap2(cw, unicw, nc);
                         rv = check(cw);
+                        if (rv && (captype == ALLCAP) && is_keepcase(rv)) rv = NULL;
                      }
-                     if ((abbv) && !(rv)) {
+                     if (abbv && !rv) {
                          *(wspace+wl) = '.';
                          *(wspace+wl+1) = '\0';
                          rv = check(wspace);
-			 if (!rv) {
+			 if (!rv || is_keepcase(rv)) {
 		            memcpy(wspace, cw, wl2);
                             *(wspace+wl2) = '.';
                             *(wspace+wl2+1) = '\0';
 			    rv = check(wspace);
+                            if (rv && ((captype == ALLCAP) && is_keepcase(rv))) rv = NULL;
 			 }
                      }
 		     break;
@@ -436,8 +435,8 @@ int Hunspell::spell(const char * word)
   
   if (rv) return 1;
 
-  // LANG_hu, LANG_de section: compoundings with dashes
-  if ((langnum == LANG_hu) or (langnum == LANG_de)) {
+  // LANG_hu: compoundings with dashes XXX deprecated!
+  if (langnum == LANG_hu) {
     int n;
     // compound word with dash (HU) I18n
     char * dash;
@@ -480,104 +479,125 @@ int Hunspell::spell(const char * word)
   return 0;
 }
 
-struct hentry * Hunspell::check(const char * word)
+struct hentry * Hunspell::check(const char * w)
 {
   struct hentry * he = NULL;
   int len;
+  char w2[MAXWORDUTF8LEN];
+  const char * word = w;
 
-  forbidden_compound = 0;
-  prevcompound = 0;
-  prevroot = NULL;
-    
-  if (pHMgr)
-     he = pHMgr->lookup(word);
+  // word reversing wrapper for complex prefixes
+  if (complexprefixes) {
+    strcpy(w2, w);
+    if (utf8) reverseword_utf(w2); else reverseword(w2);
+    word = w2;
+  }
 
-    // BUG: homonyms, not threadsafe     
-    // check forbidden words
+  forbidden_compound = 0; // XXX LANG_hu class variable for suggestions (not threadsafe)
+  prevcompound = 0;       // compounding information for Hunspell's pipe interface (not threadsafe)
+  prevroot = NULL;        // root information for Hunspell's pipe interface (not threadsafe)
+
+  // look word in hash table
+  if (pHMgr) he = pHMgr->lookup(word);
+
+  // check forbidden and onlyincompound words
   if ((he) && (he->astr) && (pAMgr) && TESTAFF(he->astr, pAMgr->get_forbiddenword(), he->alen)) {
+    // LANG_hu section: set dash information for suggestions
+    if (langnum == LANG_hu) {
     	forbidden_compound = 1;
 	if (pAMgr->get_compoundflag() &&
             TESTAFF(he->astr, pAMgr->get_compoundflag(), he->alen)) {
 		forbidden_compound = 2;
 	}
-	return NULL;
+    }
+    return NULL;
   }
 
-  if (((he == NULL) || TESTAFF(he->astr, pAMgr->get_pseudoroot(), he->alen)) && (pAMgr)) {
+  // he = next not pseudoroot and not onlyincompound homonym or NULL
+  while (he && (he->astr) &&
+    ((pAMgr->get_pseudoroot() && TESTAFF(he->astr, pAMgr->get_pseudoroot(), he->alen)) ||
+       (pAMgr->get_onlyincompound() && TESTAFF(he->astr, pAMgr->get_onlyincompound(), he->alen))
+    )) he = he->next_homonym;
+
+  // check with affixes
+  if (!he && pAMgr) {
      // try stripping off affixes */
      len = strlen(word);
      he = pAMgr->affix_check(word, len, 0);
 
+     // check compound restriction
+     if (he && he->astr && pAMgr->get_onlyincompound() &&
+         TESTAFF(he->astr, pAMgr->get_onlyincompound(), he->alen)) he = NULL;
+
      // try check compound word
      if (he) {
-        if ((he->astr) && (pAMgr) &&
-	    TESTAFF(he->astr, pAMgr->get_forbiddenword(), he->alen)
-	    && (! TESTAFF(he->astr, pAMgr->get_onlyroot(), he->alen))) {
-		forbidden_compound = 1;
-		return NULL;
-	    }
+        if ((he->astr) && (pAMgr) && TESTAFF(he->astr, pAMgr->get_forbiddenword(), he->alen)) {
+	    forbidden_compound = 1; // LANG_hu
+	    return NULL;
+        }
 	prevroot = he->word;
-     } else if (pAMgr->get_compoundflag() || pAMgr->get_compoundbegin()) {
+     } else if (pAMgr->get_compound()) {
           he = pAMgr->compound_check(word, len, 
-	                          0,0,100,0,NULL,NULL);
-    // LANG_hu section: `moving rule' with last dash
+	                          0,0,100,0,NULL,0,NULL,NULL,0);
+          // LANG_hu section: `moving rule' with last dash
 	  if ((!he) && (langnum == LANG_hu) && (word[len-1]=='-')) {
 	     char * dup = mystrdup(word);
 	     dup[len-1] = '\0';
              he = pAMgr->compound_check(dup, len-1, 
-	                          -5,0,100,1,NULL,NULL);
+	                          -5,0,100,0,NULL,1,NULL,NULL,0);
 	     free(dup);
 	  }
-    // end of LANG speficic region          
+          // end of LANG speficic region          
 	  if (he) {
-		prevroot = he->word; // XXX
+		prevroot = he->word;
 		prevcompound = 1;
 	  }
      }
 
   }
 
-  if (he) return he;
-  
-  return NULL;
+  return he;
 }
 
 int Hunspell::suggest(char*** slst, const char * word)
 {
-  char cw[MAXWORDLEN+1];
-  char wspace[MAXWORDLEN+1];
+  char cw[MAXWORDUTF8LEN + 4];
+  char wspace[MAXWORDUTF8LEN + 4];
   if (! pSMgr) return 0;
-  w_char unicw[MAXWORDLEN];
+  w_char unicw[MAXWORDLEN + 1];
   int nc = strlen(word);
-  if (nc > (MAXWORDLEN-1) / 3) return 0;
+  if (utf8) {
+    if (nc >= MAXWORDUTF8LEN) return 0;
+  } else {
+    if (nc >= MAXWORDLEN) return 0;
+  }
   int captype = 0;
   int abbv = 0;
   int wl = cleanword2(cw, word, unicw, &nc, &captype, &abbv);
   if (wl == 0) return 0;
   int ns = 0;
-  *slst = NULL; // HU, nsug in pSMgr->suggest
-  
+  *slst = NULL;
+  int capwords = 0;
+  int ngramsugs = 0;
+
   switch(captype) {
      case NOCAP:   { 
                      ns = pSMgr->suggest(slst, cw, ns);
-                     if (ns>0) break;
                      break;
                    }
 
      case INITCAP: { 
+                     capwords = 1;
+                     ns = pSMgr->suggest(slst, cw, ns);
+                     if (ns == -1) break;
                      memcpy(wspace,cw,(wl+1));
                      mkallsmall2(wspace, unicw, nc);
                      ns = pSMgr->suggest(slst, wspace, ns);
-                     for (int j=0; j < ns; j++)
-                       mkinitcap((*slst)[j]);
-	                   ns = pSMgr->suggest(slst, cw, ns);
                      break;
-		     
                    }
-
      case HUHCAP: { 
                      ns = pSMgr->suggest(slst, cw, ns);
-                     if (ns == 0) {
+                     if (ns != -1) {
                         memcpy(wspace,cw,(wl+1));
                         mkallsmall2(wspace, unicw, nc);
                         ns = pSMgr->suggest(slst, wspace, ns);
@@ -589,6 +609,7 @@ int Hunspell::suggest(char*** slst, const char * word)
                      memcpy(wspace, cw, (wl+1));
                      mkallsmall2(wspace, unicw, nc);
                      ns = pSMgr->suggest(slst, wspace, ns);
+                     if (ns == -1) break;
                      mkinitcap2(wspace, unicw, nc);
                      ns = pSMgr->suggest(slst, wspace, ns);
                      for (int j=0; j < ns; j++) {
@@ -621,47 +642,124 @@ int Hunspell::suggest(char*** slst, const char * word)
           if (pos) *pos = ' ';
       }
   }
+  // END OF LANG_de section
 
   // try ngram approach since found nothing
-  if (ns == 0) { 
-     ns = pSMgr->ngsuggest(*slst, cw, pHMgr);
-     if (ns) {
-         switch(captype) {
-	    case NOCAP:  break;
-            case HUHCAP: break; 
-            case INITCAP: { 
-                            for (int j=0; j < ns; j++)
-                              mkinitcap((*slst)[j]);
-                          }
-                          break;
-
-            case ALLCAP: { 
-                            for (int j=0; j < ns; j++)
-                              mkallcap((*slst)[j]);
-                         } 
-                         break;
+  if ((ns == 0) && pAMgr && (pAMgr->get_maxngramsugs() != 0)) {
+      ngramsugs = 1;
+      switch(captype) {
+	  case NOCAP: {
+              ns = pSMgr->ngsuggest(*slst, cw, pHMgr);
+              break;
+          }
+          case HUHCAP: {
+              memcpy(wspace,cw,(wl+1));
+              mkallsmall2(wspace, unicw, nc);
+              ns = pSMgr->ngsuggest(*slst, wspace, pHMgr);
+              break;
+          }
+          case INITCAP: { 
+              capwords = 1;
+              memcpy(wspace,cw,(wl+1));
+              mkallsmall2(wspace, unicw, nc);
+              ns = pSMgr->ngsuggest(*slst, wspace, pHMgr);
+              break;
+          }
+          case ALLCAP: {
+              memcpy(wspace,cw,(wl+1));
+              mkallsmall2(wspace, unicw, nc);
+              ns = pSMgr->ngsuggest(*slst, wspace, pHMgr);
+              for (int j=0; j < ns; j++) 
+                  mkallcap((*slst)[j]);
+              break;
 	 }
-     }
+      }
+  }
+
+  // word reversing wrapper for complex prefixes
+  if (complexprefixes) {
+    for (int j = 0; j < ns; j++) {
+      if (utf8) reverseword_utf((*slst)[j]); else reverseword((*slst)[j]);
+    }
+  }
+
+  // capitalize and erase capitalized duplications
+  if (capwords) {
+    int l = 0;
+    for (int j=0; j < ns; j++) {
+      mkinitcap((*slst)[j]);
+      (*slst)[l] = (*slst)[j];
+      for (int k=0; k < l; k++) {
+        if (strcmp((*slst)[k], (*slst)[j]) == 0) {
+          free((*slst)[j]);
+          l--;
+        }
+      }
+      l++;
+    }
+    ns = l;
   }
 
   // expand suggestions with dot(s)
-  if (abbv) {
-      for (int j = 0; j < ns; j++) {
-          (*slst)[j] = (char *) realloc((*slst)[j], strlen((*slst)[j]) + 1 + abbv);
-          strcat((*slst)[j], word + strlen(word) - abbv);
+  if (abbv && pAMgr && pAMgr->get_sugswithdots()) {
+    for (int j = 0; j < ns; j++) {
+      (*slst)[j] = (char *) realloc((*slst)[j], strlen((*slst)[j]) + 1 + abbv);
+      strcat((*slst)[j], word + strlen(word) - abbv);
+    }
+  }
+
+  // suggest keepcase
+  switch (captype) {
+    case INITCAP:
+    case ALLCAP: {
+      int l = 0;
+      for (int j=0; j < ns; j++) {
+        if (!spell((*slst)[j])) {
+          int len;
+          char s[MAXSWUTF8L];
+          w_char w[MAXSWL];
+          if (utf8) {
+            len = u8_u16(w, MAXSWL, (*slst)[j]);
+          } else {
+            strcpy(s, (*slst)[j]);
+            len = strlen(s);
+          }
+          int wl = mkallsmall2(s, w, len);
+          free((*slst)[j]);          
+          if (spell(s)) {
+            (*slst)[j] = mystrdup(s);
+            l++;
+          } else {
+            int wl = mkinitcap2(s, w, len);
+            if (spell(s)) {
+              (*slst)[j] = mystrdup(s);
+              l++;
+            }
+          }
+        } else {
+          (*slst)[l] = (*slst)[j];
+          l++;
+        }
       }
+      ns = l;
+    }
   }
 
   return ns;
 }
 
+// XXX need UTF-8 support
 int Hunspell::suggest_auto(char*** slst, const char * word)
 {
-  char cw[MAXWORDLEN+1];
-  char wspace[MAXWORDLEN+1];
+  char cw[MAXWORDUTF8LEN + 4];
+  char wspace[MAXWORDUTF8LEN + 4];
   if (! pSMgr) return 0;
   int wl = strlen(word);
-  if (wl > (MAXWORDLEN-1)) return 0;
+  if (utf8) {
+    if (wl >= MAXWORDUTF8LEN) return 0;
+  } else {
+    if (wl >= MAXWORDLEN) return 0;
+  }
   int captype = 0;
   int abbv = 0;
   wl = cleanword(cw, word, &captype, &abbv);
@@ -711,16 +809,20 @@ int Hunspell::suggest_auto(char*** slst, const char * word)
                    }
   }
 
-  // expand suggestions with dot(s)
-    if (abbv) {
-	for (int j = 0; j < ns; j++) {
-	    char * l = (char *) malloc(strlen((*slst)[j]) + 1 + abbv);
-	    strcpy(l, (*slst)[j]);
-	    strcat(l, word + strlen(word) - abbv);
-	    free((*slst)[j]);
-	    (*slst)[j] = l;			
-	}
+  // word reversing wrapper for complex prefixes
+  if (complexprefixes) {
+    for (int j = 0; j < ns; j++) {
+      if (utf8) reverseword_utf((*slst)[j]); else reverseword((*slst)[j]);
     }
+  }
+
+  // expand suggestions with dot(s)
+  if (abbv && pAMgr && pAMgr->get_sugswithdots()) {
+    for (int j = 0; j < ns; j++) {
+      (*slst)[j] = (char *) realloc((*slst)[j], strlen((*slst)[j]) + 1 + abbv);
+      strcat((*slst)[j], word + strlen(word) - abbv);
+    }
+  }
 
   // replace '-' with ' '
   if (forbidden_compound == 2) {
@@ -732,14 +834,18 @@ int Hunspell::suggest_auto(char*** slst, const char * word)
   return ns;
 }
 
-
+// XXX need UTF-8 support
 int Hunspell::stem(char*** slst, const char * word)
 {
-  char cw[MAXWORDLEN+1];
-  char wspace[MAXWORDLEN+1];
+  char cw[MAXWORDUTF8LEN + 4];
+  char wspace[MAXWORDUTF8LEN + 4];
   if (! pSMgr) return 0;
   int wl = strlen(word);
-  if (wl > (MAXWORDLEN-1)) return 0;
+  if (utf8) {
+    if (wl >= MAXWORDUTF8LEN) return 0;
+  } else {
+    if (wl >= MAXWORDLEN) return 0;
+  }
   int captype = 0;
   int abbv = 0;
   wl = cleanword(cw, word, &captype, &abbv);
@@ -818,11 +924,15 @@ int Hunspell::stem(char*** slst, const char * word)
 
 int Hunspell::suggest_pos_stems(char*** slst, const char * word)
 {
-  char cw[MAXWORDLEN+1];
-  char wspace[MAXWORDLEN+1];
+  char cw[MAXWORDUTF8LEN + 4];
+  char wspace[MAXWORDUTF8LEN + 4];
   if (! pSMgr) return 0;
   int wl = strlen(word);
-  if (wl > (MAXWORDLEN-1)) return 0;
+  if (utf8) {
+    if (wl >= MAXWORDUTF8LEN) return 0;
+  } else {
+    if (wl >= MAXWORDLEN) return 0;
+  }
   int captype = 0;
   int abbv = 0;
   wl = cleanword(cw, word, &captype, &abbv);
@@ -916,7 +1026,7 @@ void Hunspell::mkinitcap(char * p)
       unsigned short i = utfconv[(u[0].h << 8) + u[0].l].cupper;
       u[0].h = (unsigned char) (i >> 8);
       u[0].l = (unsigned char) (i & 0x00FF);
-      u16_u8(p, MAXWORDLEN, u, len);
+      u16_u8(p, MAXWORDUTF8LEN, u, len);
   }
 }
 
@@ -928,7 +1038,7 @@ int Hunspell::mkinitcap2(char * p, w_char * u, int nc)
       unsigned short i = utfconv[(u[0].h << 8) + u[0].l].cupper;
       u[0].h = (unsigned char) (i >> 8);
       u[0].l = (unsigned char) (i & 0x00FF);
-      u16_u8(p, MAXWORDLEN, u, nc);
+      u16_u8(p, MAXWORDUTF8LEN, u, nc);
       return strlen(p);
   }
   return nc;
@@ -968,18 +1078,28 @@ const char * Hunspell::get_version()
   return pAMgr->get_version();
 }
 
-
+// XXX need UTF-8 support
 char * Hunspell::morph(const char * word)
 {
-  char cw[MAXWORDLEN+1];
-  char wspace[MAXWORDLEN+1];
+  char cw[MAXWORDUTF8LEN + 4];
+  char wspace[MAXWORDUTF8LEN + 4];
   if (! pSMgr) return 0;
   int wl = strlen(word);
-  if (wl > (MAXWORDLEN-1)) return 0;
+  if (utf8) {
+    if (wl >= MAXWORDUTF8LEN) return 0;
+  } else {
+    if (wl >= MAXWORDLEN) return 0;
+  }
   int captype = 0;
   int abbv = 0;
   wl = cleanword(cw, word, &captype, &abbv);
-  if (wl == 0) return 0;
+  if (wl == 0) {
+      if (abbv) {
+          for (wl = 0; wl < abbv; wl++) cw[wl] = '.';
+          cw[wl] = '\0';
+          abbv = 0;
+      } else return 0;
+  }
 
   char result[MAXLNLEN];
   char * st = NULL;
@@ -990,8 +1110,9 @@ char * Hunspell::morph(const char * word)
   int n2 = 0;
   int n3 = 0;
 
-  // test numbers and 56%-kal, 45%-kal 50-50%, 11.20-kor, but
-  // numbers with two, or more dots are bad (for example 2003.07.22.)
+  // test numbers
+  // LANG_hu section: set dash information for suggestions
+  if (langnum == LANG_hu) {
   while ((n < wl) && 
   	(((cw[n] <= '9') && (cw[n] >= '0')) || (((cw[n] == '.') || (cw[n] == ',')) && (n > 0)))) {
   	n++;
@@ -1003,7 +1124,6 @@ char * Hunspell::morph(const char * word)
 	}
   }
 
-  // numbers with 3, or more digits (HU)
   if ((n == wl) && (n3 > 0) && (n - n3 > 3)) return NULL;
   if ((n == wl) || ((n>0) && ((cw[n]=='%') || (cw[n]=='')) && check(cw+n))) {
   	strcat(result, cw);
@@ -1032,7 +1152,8 @@ char * Hunspell::morph(const char * word)
 	}
 	return mystrdup(result);
   }
-  //  if ((cw[n]=='.') && check(cw + n + 1)) return 1;
+  }
+  // END OF LANG_de section
   
   switch(captype) {
      case NOCAP:   { 
@@ -1155,12 +1276,19 @@ char * Hunspell::morph(const char * word)
                    }
   }
 
-  if (result && (*result)) return mystrdup(result);
+  if (result && (*result)) {
+    // word reversing wrapper for complex prefixes
+    if (complexprefixes) {
+      if (utf8) reverseword_utf(result); else reverseword(result);
+    }
+    return mystrdup(result);
+  }
 
   // compound word with dash (HU) I18n
   char * dash;
   int nresult = 0;
-  if ((dash=(char *) strchr(cw,'-'))) {
+  // LANG_hu section: set dash information for suggestions
+  if ((langnum == LANG_hu) && (dash=(char *) strchr(cw,'-'))) {
       *dash='\0';      
       // examine 2 sides of the dash
       if (dash[1] == '\0') { // base word ending with dash
@@ -1235,13 +1363,18 @@ char * Hunspell::morph(const char * word)
   return NULL;
 }
 
+// XXX need UTF-8 support
 char * Hunspell::morph_with_correction(const char * word)
 {
-  char cw[MAXWORDLEN+1];
-  char wspace[MAXWORDLEN+1];
+  char cw[MAXWORDUTF8LEN + 4];
+  char wspace[MAXWORDUTF8LEN + 4];
   if (! pSMgr) return 0;
   int wl = strlen(word);
-  if (wl > (MAXWORDLEN-1)) return 0;
+  if (utf8) {
+    if (wl >= MAXWORDUTF8LEN) return 0;
+  } else {
+    if (wl >= MAXWORDLEN) return 0;
+  }
   int captype = 0;
   int abbv = 0;
   wl = cleanword(cw, word, &captype, &abbv);
@@ -1403,3 +1536,4 @@ int Hunspell::analyze(char ***out, const char *word) {
   free(m);
   return n;
 }
+
